DEFAULT = {
    "cardholder_sync": False,
    "user_sync_field": "CardNumber",
    "nemo_user_sync_field": "badge_number",
    # 601 is access granted, 607 Access Granted Door Unlocked
    "relevant_event_types": [601, 607],
    # we will store the date to restart as the last processed event minus this buffer
    # this allows us to make sure we are not missing events between calls to all the separate readers
    "buffer_in_seconds": 7,
    # whether to record someone attempting to log into an area they are already logged in as error in logs
    "record_already_logged_in_as_error": True,
}

RS2_ACCESS = {}



# Generated by Django 3.2.18 on 2023-05-26 20:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO_rs2_access', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Cardholder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cardholder_id', models.CharField(max_length=250, unique=True)),
                ('cardholder_name', models.CharField(blank=True, max_length=250, null=True)),
                ('key_name', models.CharField(max_length=250)),
                ('key_value', models.CharField(max_length=250)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



from argparse import Namespace
from datetime import datetime, timedelta
from typing import Dict, List

from NEMO.models import Area
from NEMO.utilities import export_format_datetime, quiet_int
from django.utils import timezone
from drf_excel.mixins import XLSXFileMixin
from rest_framework import status
from rest_framework.fields import CharField, DateTimeField
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.viewsets import GenericViewSet

from NEMO_rs2_access.models import Reader
from NEMO_rs2_access.rs2 import get_cardholder_property, get_sorted_reader_events


class LastEntriesSerializer(Serializer):
    cardholder_name = CharField(read_only=True)
    cardholder_id = CharField(read_only=True)
    cardholder_user_text1 = CharField(read_only=True)
    card_number = CharField(read_only=True)
    card_id = CharField(read_only=True)
    login_time = DateTimeField(read_only=True)
    reader_id = CharField(read_only=True)
    reader_location = CharField(read_only=True)
    area_name = CharField(read_only=True)

    def update(self, instance, validated_data):
        pass

    def create(self, validated_data):
        pass

    class Meta:
        fields = "__all__"


class LastEntriesViewSet(XLSXFileMixin, GenericViewSet):
    serializer_class = LastEntriesSerializer
    since_minutes = None

    def check_permissions(self, request):
        pass

    def list(self, request, *args, **kwargs):
        self.since_minutes = quiet_int(self.request.GET.get("last", None), None)
        if self.since_minutes is None:
            return Response(
                status=status.HTTP_400_BAD_REQUEST,
                data={
                    "last": "This field is required. Please specify the number of minutes to check last users entries for"
                },
            )
        try:
            queryset = self.get_queryset()
        except Exception as e:
            return Response(status=status.HTTP_400_BAD_REQUEST, data=str(e))
        serializer = self.serializer_class(queryset, many=True)
        return Response(serializer.data)

    def get_queryset(self):
        since = timezone.now() - timedelta(minutes=self.since_minutes)

        last_entries = get_last_entries(since)
        last_entries.sort(key=lambda x: x.login_time, reverse=True)
        return last_entries

    def get_filename(self, *args, **kwargs):
        return f"last_entries-{self.since_minutes}.xlsx"


def get_last_entries(since: datetime) -> List:
    last_entries = []
    formatted_since = since.astimezone().replace(tzinfo=None).isoformat(sep="T", timespec="seconds")
    events, readers, cardholders = get_sorted_reader_events(formatted_since, [])
    # cache the text1 value for cardholders
    cardholder_text1_cache = {}
    events_dict: Dict[str, Dict] = {}
    # Put event into a dict by area id, this way the latest event for that area will
    # override the previous one, and we only care about the last entry/exit per area
    for event in events:
        area: Area = readers.get(event["SourceId"]).area
        if area:
            event_date = event["EventDateObject"]
            auto_logout = getattr(area, "auto_logout_time", None)
            # remove events that should have been auto logged out
            if not auto_logout or (event_date + timedelta(minutes=auto_logout) > timezone.now()):
                cardholder_id = event["CardholderId"]
                event["area_name"] = area.name
                text1 = cardholder_text1_cache.get(cardholder_id) or cardholder_text1_cache.setdefault(
                    cardholder_id, get_cardholder_property(cardholder_id, "UserColumns.UserText1", save=False)
                )
                event["user_columns_user_text1"] = text1
                user_location_id = f"{area.id}-{cardholder_id}"
                events_dict[user_location_id] = event
    for event in events_dict.values():
        # Only keep entrance reader types
        if readers.get(event["SourceId"]).reader_type == Reader.ReaderType.ENTRANCE:
            last_entries.append(
                Namespace(
                    **{
                        "cardholder_name": event["Cardholder"].strip(),
                        "cardholder_id": event["CardholderId"],
                        "cardholder_user_text1": event["user_columns_user_text1"],
                        "card_number": event["CardNumber"],
                        "card_id": event["CardId"],
                        "login_time": event["EventDate"],
                        "reader_id": event["SourceId"],
                        "reader_location": event["EventLocation"],
                        "area_name": event["area_name"],
                    }
                )
            )
    return last_entries



from django.core.management import BaseCommand

from NEMO_rs2_access import rs2


class Command(BaseCommand):
    help = "Run every few minutes to update the project to charge on current area access records."

    def handle(self, *args, **options):
        rs2.update_current_access_project()



import json

from NEMO.typing import QuerySetType
from NEMO.utilities import BasicDisplayTable, export_format_datetime
from django import forms
from django.contrib import admin
from django.template.defaultfilters import yesno
from django.utils.safestring import mark_safe

from NEMO_rs2_access.models import Cardholder, ErrorLog, Reader, UserPreferencesDefaultProject
from NEMO_rs2_access.rs2 import sync_access, sync_readers


@admin.action(description="Sync all readers with RS2")
def admin_sync_readers(modeladmin, request, queryset):
    sync_readers()


@admin.action(description="Sync user access RS2")
def admin_sync_access(modeladmin, request, queryset):
    sync_access()


@admin.action(description="Export selected error logs")
def admin_export_error_logs(modeladmin, request, queryset: QuerySetType[ErrorLog]):
    table_result = BasicDisplayTable()
    table_result.add_header(("type", "Type"))
    table_result.add_header(("created", "Created"))
    table_result.add_header(("description", "Description"))
    table_result.add_header(("cardholder_value", "Cardholder value"))
    table_result.add_header(("cardholder_name", "Cardholder name"))
    table_result.add_header(("event_time", "Event time"))
    table_result.add_header(("reader", "Reader"))
    table_result.add_header(("fixed", "Fixed"))
    for error_log in queryset:
        table_result.add_row(
            {
                "type": error_log.get_error_type_display(),
                "created": error_log.created,
                "description": error_log.description,
                "cardholder_value": error_log.cardholder_value,
                "cardholder_name": error_log.get_cardholder_name(),
                "event_time": error_log.get_event_time(),
                "reader": error_log.get_reader(),
                "fixed": error_log.fixed(),
            }
        )

    filename = f"error_logs_{export_format_datetime()}.csv"
    response = table_result.to_csv()
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    return response


class UserPreferencesDefaultProjectAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.user_preferences_id:
            self.fields["user_preferences"].disabled = True
            self.fields["default_project"].queryset = self.instance.user_preferences.user.active_projects()

    class Meta:
        model = UserPreferencesDefaultProject
        fields = "__all__"


@admin.register(UserPreferencesDefaultProject)
class UserPreferencesDefaultProjectAdmin(admin.ModelAdmin):
    list_display = ["get_user", "default_project"]
    list_filter = (("default_project", admin.RelatedOnlyFieldListFilter),)
    search_fields = [
        "user_preferences__user__first_name",
        "user_preferences__user__last_name",
        "user_preferences__user__username",
    ]
    form = UserPreferencesDefaultProjectAdminForm

    @admin.display(description="User", ordering="user_preferences__user")
    def get_user(self, obj: UserPreferencesDefaultProject):
        return obj.user_preferences.user


@admin.register(Reader)
class ReaderAdmin(admin.ModelAdmin):
    list_display = ["reader_id", "reader_name", "site_id", "area", "reader_type", "installed", "last_updated"]
    list_filter = [
        ("area", admin.RelatedOnlyFieldListFilter),
        "reader_type",
        "installed",
        "site_id",
    ]
    readonly_fields = ["reader_id", "reader_name", "site_id", "installed", "get_data", "created", "last_updated"]
    exclude = ["data"]
    search_fields = ["reader_name", "area__name"]
    actions = [admin_sync_readers, admin_sync_access]

    @admin.display(description="Data", ordering="data")
    def get_data(self, obj: Reader):
        result = json.dumps(obj.data, indent=4, sort_keys=True)
        result_str = f"<pre>{result}</pre>"
        return mark_safe(result_str)


@admin.register(Cardholder)
class CardholderAdmin(admin.ModelAdmin):
    list_display = ["cardholder_id", "cardholder_name", "key_name", "key_value", "created", "last_updated"]
    readonly_fields = ["created", "last_updated"]


@admin.register(ErrorLog)
class ErrorLogAdmin(admin.ModelAdmin):
    list_display = [
        "created",
        "get_error_type_display",
        "get_event_time",
        "get_reader",
        "cardholder_value",
        "get_cardholder_name",
        "user",
        "appears_fixed",
    ]
    readonly_fields = ["event_data"]
    list_filter = ["error_type"]
    date_hierarchy = "created"
    actions = [admin_export_error_logs]

    @admin.display(description="Appears fixed")
    def appears_fixed(self, error_log: ErrorLog) -> str:
        return yesno(error_log.fixed(), "Yes,No,")



from NEMO.urls import router
from django.urls import path

from NEMO_rs2_access import api, views

router.register(r"rs2/last_entries", api.LastEntriesViewSet, basename="last_entries")
router.registry.sort(key=lambda x: (x[0].count("/"), x[0]))

urlpatterns = [
    # Override user preferences to add default project selection
    path("user_preferences/", views.custom_user_preferences, name="user_preferences"),
    path("rs2_sync_readers/", views.rs2_sync_reader, name="rs2_sync_readers"),
    path("rs2_sync_access/", views.rs2_sync_access, name="rs2_sync_access"),
    path(
        "rs2_update_current_access_project/",
        views.rs2_update_current_access_project,
        name="rs2_update_current_access_project",
    ),
]



from django.core.management import BaseCommand

from NEMO_rs2_access import rs2


class Command(BaseCommand):
    help = "Run every day to sync up rs2 readers."

    def handle(self, *args, **options):
        rs2.sync_readers()



from logging import getLogger
from typing import Dict, Type

from NEMO.utilities import distinct_qs_value_list
from django.db.models import Model

sync_logger = getLogger(__name__)


def sync_model_rest_data(data, model_class: Type[Model], remote_id_field_name: str, mapping: Dict):
    model_id_field = f"{model_class._meta.model_name}_id"
    try:
        db_data_ids = distinct_qs_value_list(model_class.objects.all(), model_id_field)
        remote_data_ids = set()
        for item in data:
            item_id = item.get(remote_id_field_name)
            remote_data_ids.add(item_id)
            defaults = {model_field: item.get(remote_field) for model_field, remote_field in mapping.items()}
            defaults["data"] = item
            model_class.objects.update_or_create(**{model_id_field: item_id}, defaults=defaults)
        to_remove = db_data_ids.difference(remote_data_ids)
        model_class.objects.filter(**{f"{model_id_field}__in": to_remove}).delete()
        if to_remove:
            sync_logger.info(f"Deleted {model_class} with ids: {', '.join(str(x) for x in to_remove)}")
    except:
        sync_logger.exception(f"Error syncing {model_class} table")



import json
from datetime import datetime
from typing import Optional

from NEMO.models import Area, Project, User, UserPreferences
from NEMO.utilities import format_datetime
from NEMO.views.constants import CHAR_FIELD_MAXIMUM_LENGTH
from django.db import models


class UserPreferencesDefaultProject(models.Model):
    user_preferences = models.OneToOneField(UserPreferences, on_delete=models.CASCADE, related_name="default_project")
    default_project = models.ForeignKey(Project, null=True, blank=True, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.user_preferences.user) + " - " + str(self.default_project or "")

    class Meta:
        verbose_name = "Default project"


class Reader(models.Model):
    class ReaderType(object):
        ENTRANCE = 1
        EXIT = 2
        Choices = (
            (ENTRANCE, "Entrance"),
            (EXIT, "Exit"),
        )

    reader_id = models.CharField(max_length=250, unique=True)
    site_id = models.CharField(max_length=250)
    reader_name = models.CharField(max_length=250)
    installed = models.BooleanField()
    data = models.JSONField()
    area = models.ForeignKey(Area, null=True, blank=True, on_delete=models.CASCADE)
    reader_type = models.PositiveIntegerField(null=True, choices=ReaderType.Choices)
    last_updated = models.DateTimeField(auto_now=True)
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return self.reader_name


class Cardholder(models.Model):
    cardholder_id = models.CharField(max_length=250, unique=True)
    cardholder_name = models.CharField(max_length=250, blank=True, null=True)
    key_name = models.CharField(max_length=250)
    key_value = models.CharField(max_length=250)
    last_updated = models.DateTimeField(auto_now=True)
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return self.cardholder_id


class ErrorLog(models.Model):
    class ErrorType(object):
        GENERAL = 1
        NO_ACTIVE_PROJECTS = 2
        NO_MATCHING_USER = 3
        LOGOUT_WITHOUT_RECORD = 4
        ALREADY_LOGGED_IN = 5
        Choices = (
            (GENERAL, "General"),
            (NO_ACTIVE_PROJECTS, "No active projects"),
            (NO_MATCHING_USER, "No matching user"),
            (LOGOUT_WITHOUT_RECORD, "Logout without record"),
            (ALREADY_LOGGED_IN, "Login attempt while already logged in"),
        )

    created = models.DateTimeField(auto_now_add=True)
    description = models.TextField()
    error_type = models.PositiveIntegerField(default=ErrorType.GENERAL, choices=ErrorType.Choices)
    cardholder_value = models.CharField(null=True, blank=True, max_length=CHAR_FIELD_MAXIMUM_LENGTH)
    user = models.ForeignKey(User, null=True, blank=True, on_delete=models.SET_NULL)
    event_data = models.TextField(null=True, blank=True)

    def fixed(self):
        if self.error_type == ErrorLog.ErrorType.NO_ACTIVE_PROJECTS:
            try:
                if self.user:
                    return self.user.active_projects().exists()
            except User.DoesNotExist:
                pass
        elif self.error_type == ErrorLog.ErrorType.NO_MATCHING_USER:
            try:
                from NEMO_rs2_access.rs2 import find_user

                return find_user(self.cardholder_value) is not None
            except:
                pass

    def get_reader(self) -> Optional[Reader]:
        if self.event_data:
            try:
                data = json.loads(self.event_data)
                return Reader.objects.get(reader_id=data["SourceId"])
            except:
                pass

    def get_cardholder_name(self) -> Optional[str]:
        if self.event_data:
            try:
                data = json.loads(self.event_data)
                return data["Cardholder"]
            except:
                pass

    def get_event_time(self) -> Optional[datetime]:
        if self.event_data:
            try:
                data = json.loads(self.event_data)
                return datetime.fromisoformat(data["EventDate"]).astimezone()
            except:
                pass

    def __str__(self):
        return f"{self.get_error_type_display()} error on {format_datetime(self.created)}"

    class Meta:
        ordering = ["-created"]



# Generated by Django 3.2.18 on 2023-05-18 21:57

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('NEMO', '0045_version_4_5_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserPreferencesDefaultProject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('default_project', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.project')),
                ('user_preferences', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='default_project', to='NEMO.userpreferences')),
            ],
            options={
                'verbose_name': 'Default project',
            },
        ),
        migrations.CreateModel(
            name='Reader',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('reader_id', models.CharField(max_length=250, unique=True)),
                ('site_id', models.CharField(max_length=250)),
                ('reader_name', models.CharField(max_length=250)),
                ('installed', models.BooleanField()),
                ('data', models.JSONField()),
                ('reader_type', models.PositiveIntegerField(choices=[(1, 'Entrance'), (2, 'Exit')], null=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('area', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.area')),
            ],
        ),
    ]



# Generated by Django 4.2.16 on 2024-11-19 23:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_rs2_access", "0004_errorlog_event_data"),
    ]

    operations = [
        migrations.AlterField(
            model_name="errorlog",
            name="error_type",
            field=models.PositiveIntegerField(
                choices=[
                    (1, "General"),
                    (2, "No active projects"),
                    (3, "No matching user"),
                    (4, "Logout without record"),
                    (5, "Login attempt while already logged in"),
                ],
                default=1,
            ),
        ),
    ]



from copy import copy
from datetime import timedelta
from typing import Dict, List
from unittest import mock

from NEMO.models import Account, Area, AreaAccessRecord, Customization, Project, Tool, UsageEvent, User
from django.conf import settings
from django.test import TestCase
from django.utils import timezone
from requests import Response

from NEMO_rs2_access.models import ErrorLog, Reader, UserPreferencesDefaultProject
from NEMO_rs2_access.rs2 import (
    LAST_EVENTS_PROCESSED,
    NEXT_EVENTS_BEGIN_DATE,
    READERS_HISTORY_URL,
    get_project_ids_to_charge_for_user,
    get_rs2_settings,
    login,
    sync_access,
)

READER_ONE_ID = "123-456-789"
READER_TWO_ID = "456-789-101"

BADGE_NUMBER = "123"


# This method will be used by the mock to replace requests.get
# In the web relay case, it will return a xml file with the relay statuses
def mocked_requests_get(*args, **kwargs):
    rs2_settings = get_rs2_settings()

    class MockResponse(Response):
        def __init__(self, json_data, status_code):
            super().__init__()
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

    reader_id = settings.RS2_READER_ID
    if args[0].startswith(rs2_settings.url + READERS_HISTORY_URL.format(reader_id, "")):
        badge = settings.RS2_BADGE_NUMBER
        return MockResponse(mock_event_access_granted(badge, reader_id), 200)
    elif args[0].startswith(rs2_settings.url):
        return MockResponse([], 200)

    return MockResponse(None, 404)


def mock_event_access_granted(badge_number, reader_id) -> List[Dict]:
    rs2_settings = get_rs2_settings()
    event = {
        "SiteId": "123123123123123",
        "SiteName": "Site 1",
        "EventDate": "2023-05-15T14:09:46",
        "SourceType": 6,
        "SourceId": reader_id,
        "EventType": 601,
        "Description": "Access Granted",
        "EventLocation": "Cleanroom entrance",
        "EventLocationId": None,
        "Cardholder": "Testy McTester",
        "CardNumber": None,
        "FacilityCode": 507,
        "CardholderId": 123456,
        "CardId": None,
        rs2_settings.user_sync_field: badge_number,
    }
    return [event]


class RS2Test(TestCase):
    def test_get_correct_project(self):
        account: Account = Account.objects.create(name="Test account")
        project: Project = Project.objects.create(name="Test project", account=account)
        project2: Project = Project.objects.create(name="Test project 2", account=account)
        project3: Project = Project.objects.create(
            name="Test project 3", account=Account.objects.create(name="Test account 2")
        )
        user, created = User.objects.get_or_create(
            username="testy", first_name="McTester", last_name="testy", badge_number=1
        )
        random_user, created = User.objects.get_or_create(
            username="randy", first_name="McRander", last_name="randy", badge_number=11
        )
        random_user.projects.add(project, project2)
        start = timezone.now() - timedelta(minutes=30)
        end = timezone.now()
        # no project, return the given project id
        self.assertEqual([1], get_project_ids_to_charge_for_user(user, start, end, 1))
        # one project, should be the one
        user.projects.add(project)
        self.assertEqual([project.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        user.projects.add(project2)
        project.active = False
        project.save()
        # two projects, first active one returned
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        project.active = True
        project.save()
        # default project set, should be the one returned
        UserPreferencesDefaultProject.objects.create(user_preferences=user.get_preferences(), default_project=project2)
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create one Usage event that is remote, no change
        tool = Tool.objects.create(name="test tool")
        UsageEvent.objects.create(
            user=user,
            project=project,
            remote_work=True,
            tool=tool,
            operator=user,
            start=(start - timedelta(hours=1)),
            end=timezone.now(),
        )
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create another, non-remote, which ended before
        UsageEvent.objects.create(
            user=user,
            project=project,
            tool=tool,
            operator=user,
            start=(start - timedelta(hours=1)),
            end=(start - timedelta(seconds=1)),
        )
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create another, non-remote, which started after
        UsageEvent.objects.create(
            user=user, project=project, tool=tool, operator=user, start=(end + timedelta(seconds=1)), end=timezone.now()
        )
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # ongoing event starting later
        usage = UsageEvent.objects.create(
            user=user, project=project, tool=tool, operator=user, start=(end + timedelta(seconds=1))
        )
        usage.delete()
        self.assertEqual([project2.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # ongoing event, should work
        UsageEvent.objects.create(user=user, project=project, tool=tool, operator=user, start=start)
        self.assertEqual([project.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create another qualifying, no duplicates
        UsageEvent.objects.create(
            user=user, project=project, tool=tool, operator=user, start=(start - timedelta(hours=1)), end=timezone.now()
        )
        self.assertEqual([project.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create a third qualifying event on different project but same account, only the first one should come back
        UsageEvent.objects.create(
            user=user,
            project=project2,
            tool=tool,
            operator=user,
            start=(start - timedelta(hours=1)),
            end=timezone.now(),
        )
        self.assertEqual([project.id], get_project_ids_to_charge_for_user(user, start, end, 0))
        # create another third qualifying event on different project and different account, both should come back
        UsageEvent.objects.create(
            user=user,
            project=project3,
            tool=tool,
            operator=user,
            start=(start - timedelta(hours=1)),
            end=timezone.now(),
        )
        self.assertEqual([project.id, project3.id], get_project_ids_to_charge_for_user(user, start, end, None))

    def test_login(self):
        account: Account = Account.objects.create(name="Test account")
        project: Project = Project.objects.create(name="Test project", account=account)
        project2: Project = Project.objects.create(name="Test project 2", account=account)
        user, created = User.objects.get_or_create(
            username="testy", first_name="McTester", last_name="testy", badge_number=1
        )
        area = Area.objects.create(name="Area 1")
        start = timezone.now()
        login(user, area, start)
        records = AreaAccessRecord.objects.filter(customer=user, area=area)
        # No projects found for user, no log in
        self.assertFalse(records.exists())
        user.projects.add(project)
        login(user, area, start)
        # Now we should be ok
        self.assertTrue(records.exists())
        self.assertEqual(records.first().area, area)
        self.assertEqual(records.first().start, start)
        self.assertEqual(records.first().end, None)
        # login again, nothing should have changed since it's in the same area
        login(user, area, timezone.now())
        records = AreaAccessRecord.objects.filter(customer=user, area=area)
        self.assertTrue(records.exists())
        self.assertEqual(len(records), 1)
        self.assertEqual(records.first().start, start)
        self.assertEqual(records.first().end, None)
        # Login to different area, should log out of first one and log in second one
        new_area = Area.objects.create(name="Area 2")
        new_start = timezone.now()
        login(user, new_area, new_start)
        previous_records = AreaAccessRecord.objects.filter(customer=user, end__isnull=False)
        self.assertTrue(previous_records.exists())
        self.assertEqual(previous_records.first().end, new_start)
        self.assertEqual(previous_records.first().area, area)
        new_records = AreaAccessRecord.objects.filter(customer=user, area=new_area)
        self.assertTrue(new_records.exists())

    @mock.patch("NEMO_rs2_access.rs2.requests.get", side_effect=mocked_requests_get)
    def test_login_logout_event(self, mock_args):
        # Creating user, area and in/out readers
        account: Account = Account.objects.create(name="Test account")
        project: Project = Project.objects.create(name="Test project", account=account)
        user, created = User.objects.get_or_create(
            username="testy", first_name="McTester", last_name="testy", badge_number=BADGE_NUMBER
        )
        user.projects.add(project)
        cleanroom = Area.objects.create(name="Cleanroom")
        reader_enter = Reader.objects.create(
            reader_id=READER_ONE_ID,
            reader_name="Cleanroom entrance",
            area=cleanroom,
            reader_type=Reader.ReaderType.ENTRANCE,
            installed=True,
            data="",
        )
        reader_exit = Reader.objects.create(
            reader_id=READER_TWO_ID,
            reader_name="Cleanroom exit",
            area=cleanroom,
            reader_type=Reader.ReaderType.EXIT,
            installed=True,
            data="",
        )

        # Login
        with self.settings(RS2_BADGE_NUMBER=user.badge_number, RS2_READER_ID=READER_ONE_ID):
            sync_access()
        self.assertTrue(user.in_area())
        area_record = user.area_access_record()
        self.assertEqual(area_record.area, cleanroom)

        # Logout
        with self.settings(RS2_BADGE_NUMBER=user.badge_number, RS2_READER_ID=READER_TWO_ID):
            sync_access()
        self.assertFalse(user.in_area())
        area_record = AreaAccessRecord.objects.get(id=area_record.id)
        self.assertTrue(area_record.end)

    @mock.patch("NEMO_rs2_access.rs2.requests.get", side_effect=mocked_requests_get)
    def test_double_login_event(self, mock_args):
        # Creating user, area and in/out readers
        account: Account = Account.objects.create(name="Test account")
        project: Project = Project.objects.create(name="Test project", account=account)
        user, created = User.objects.get_or_create(
            username="testy", first_name="McTester", last_name="testy", badge_number=BADGE_NUMBER
        )
        user.projects.add(project)
        cleanroom = Area.objects.create(name="Cleanroom")
        reader_enter = Reader.objects.create(
            reader_id=READER_ONE_ID,
            reader_name="Cleanroom entrance",
            area=cleanroom,
            reader_type=Reader.ReaderType.ENTRANCE,
            installed=True,
            data="",
        )

        # Login
        area_record = AreaAccessRecord.objects.create(
            customer=user, area=cleanroom, start=timezone.now(), project=project
        )
        self.assertTrue(user.in_area())
        self.assertEqual(area_record.area, cleanroom)

        rs2_access_no_error_on_already_logged_in = copy(settings.RS2_ACCESS)
        rs2_access_no_error_on_already_logged_in["record_already_logged_in_as_error"] = False

        # Login again, flag off so no log
        with self.settings(
            RS2_BADGE_NUMBER=user.badge_number,
            RS2_READER_ID=READER_ONE_ID,
            RS2_ACCESS=rs2_access_no_error_on_already_logged_in,
        ):
            sync_access()
        self.assertFalse(ErrorLog.objects.exists())

        # Reset
        Customization.objects.filter(name=NEXT_EVENTS_BEGIN_DATE).delete()
        Customization.objects.filter(name=LAST_EVENTS_PROCESSED).delete()

        # Login again, flag on so it should be logged
        with self.settings(RS2_BADGE_NUMBER=user.badge_number, RS2_READER_ID=READER_ONE_ID):
            sync_access()
        self.assertTrue(ErrorLog.objects.exists())
        self.assertEqual(ErrorLog.objects.first().user, user)
        self.assertEqual(ErrorLog.objects.first().error_type, ErrorLog.ErrorType.ALREADY_LOGGED_IN)



from NEMO.decorators import replace_function
from NEMO.forms import UserPreferencesForm
from NEMO.models import User
from NEMO.utilities import render_combine_responses
from NEMO.views.customization import StatusDashboardCustomization
from NEMO.views.status_dashboard import show_staff_status
from NEMO.views.users import user_preferences
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse
from django.shortcuts import redirect
from django.utils import timezone
from django.views.decorators.http import require_GET, require_http_methods

from NEMO_rs2_access.models import UserPreferencesDefaultProject
from NEMO_rs2_access.rs2 import logout, sync_access, sync_readers, update_current_access_project


@login_required
@require_http_methods(["GET", "POST"])
def custom_user_preferences(request):
    original_response = user_preferences(request)
    user: User = User.objects.get(pk=request.user.id)
    if request.method == "POST":
        if original_form_is_valid(request, user):
            default_project = request.POST.get("default_project", "")
            UserPreferencesDefaultProject.objects.update_or_create(
                user_preferences=user.preferences, defaults={"default_project_id": default_project or None}
            )
            messages.success(request, "Your preferences have been saved")
            return redirect("user_preferences")
    user_default_project = ""
    try:
        user_default_project = user.preferences.default_project.default_project
    except UserPreferencesDefaultProject.DoesNotExist:
        pass
    dictionary = {"user_default_project": user_default_project, "projects": user.active_projects()}
    return render_combine_responses(
        request, original_response, "NEMO_rs2_access/preferences_default_project.html", dictionary
    )


def original_form_is_valid(request, user) -> bool:
    user_view_options = StatusDashboardCustomization.get("dashboard_staff_status_user_view")
    staff_view_options = StatusDashboardCustomization.get("dashboard_staff_status_staff_view")
    user_view = user_view_options if not user.is_staff else staff_view_options if not user.is_facility_manager else ""
    pref_form = UserPreferencesForm(data=request.POST or None, instance=user.preferences)
    if not show_staff_status(request) or user_view == "day":
        pref_form.fields["staff_status_view"].disabled = True
    return pref_form.is_valid()


# Replace NEMO's logout function by our own, so if someone logs out from the dashboard or calendar,
# it will proceed with the rs2 flow, figuring out which project to charge etc.
@replace_function("NEMO.views.area_access.log_out_user")
def new_logout(old_function, user: User):
    logout(user, timezone.now())


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def rs2_sync_reader(request):
    sync_readers()
    return HttpResponse()


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def rs2_sync_access(request):
    sync_access()
    return HttpResponse()


@login_required
@require_GET
@permission_required("NEMO.trigger_timed_services", raise_exception=True)
def rs2_update_current_access_project(request):
    update_current_access_project()
    return HttpResponse()



from argparse import Namespace
from copy import copy

from django.conf import settings

from NEMO_rs2_access import app_settings


def get_rs2_settings():
    rs2_settings = copy(app_settings.DEFAULT)
    rs2_settings.update(settings.RS2_ACCESS)
    return Namespace(**rs2_settings)


def get_event_id(event: dict):
    # Create our own, hopefully unique id: date_readerId_userId
    return f"{event['EventDate']}_{event['SourceId']}_{event['CardholderId']}"


def find_user(cardholder_sync_value):
    from NEMO.models import User

    rs2_settings = get_rs2_settings()
    return User.objects.filter(**{rs2_settings.nemo_user_sync_field: cardholder_sync_value}).first()


def find_in_json(element, json):
    keys = element.split(".")
    rv = json
    for key in keys:
        rv = rv[key]
    return rv



from django.core.management import BaseCommand

from NEMO_rs2_access import rs2


class Command(BaseCommand):
    help = "Run every few minutes to sync up with rs2 access."

    def handle(self, *args, **options):
        rs2.sync_access()



from django.apps import AppConfig
from django.conf import settings

from . import app_settings as defaults

# Set some app default settings
for name in dir(defaults):
    if name.isupper() and not hasattr(settings, name):
        setattr(settings, name, getattr(defaults, name))


class NemoRs2AccessConfig(AppConfig):
    name = "NEMO_rs2_access"
    verbose_name = "RS2 Access"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        """
        This code will be run when Django starts.
        """
        # Import the place where @replace_function is used otherwise it won't work for management commands
        from NEMO_rs2_access import views

        pass



# Generated by Django 3.2.20 on 2023-08-09 19:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO_rs2_access', '0002_cardholder'),
    ]

    operations = [
        migrations.CreateModel(
            name='ErrorLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('description', models.TextField()),
                ('error_type', models.PositiveIntegerField(choices=[(1, 'General'), (2, 'No active projects'), (3, 'No matching user')], default=1)),
                ('data', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options = {"ordering": ["-created"]}
        ),
    ]



import datetime
import json
from logging import getLogger
from typing import Dict, List, Optional, Set, Type, Union

import requests
from NEMO.exceptions import NEMOException
from NEMO.models import Area, AreaAccessRecord, Customization, EmailNotificationType, Project, UsageEvent, User
from NEMO.utilities import beginning_of_the_day, distinct_qs_value_list, get_full_url, new_model_copy, send_mail
from NEMO.views.calendar import shorten_reservation
from NEMO.views.customization import EmailsCustomization
from django.db.models import Model
from django.http import HttpResponseBadRequest
from django.urls import reverse
from django.utils import timezone
from requests.auth import HTTPBasicAuth

from NEMO_rs2_access.models import Cardholder, ErrorLog, Reader, UserPreferencesDefaultProject
from NEMO_rs2_access.sync import sync_model_rest_data
from NEMO_rs2_access.utilities import find_in_json, find_user, get_event_id, get_rs2_settings

rs2_logger = getLogger(__name__)

NEXT_EVENTS_BEGIN_DATE = "rs2_event_begin_date"
LAST_EVENTS_PROCESSED = "rs2_event_last_processed"

READERS_URL = "/Readers"
READERS_HISTORY_URL = READERS_URL + "/{}/EventHistory?beginDate={}"
CARDHOLDERS_URL = "/Cardholders/{}"

reader_mapping = {
    "reader_name": "ReaderName",
    "site_id": "SiteID",
    "installed": "DeviceInstalled",
}


def sync_access():
    rs2_settings = get_rs2_settings()
    buffer_in_seconds = rs2_settings.buffer_in_seconds
    try:
        # default date is beginning of the day today if not set in Customization
        default_date = beginning_of_the_day(datetime.datetime.now(), in_local_timezone=False).isoformat()
        last_event_date_read, created = Customization.objects.get_or_create(
            name=NEXT_EVENTS_BEGIN_DATE, defaults={"value": default_date}
        )
        last_event_processed, created = Customization.objects.get_or_create(
            name=LAST_EVENTS_PROCESSED, defaults={"value": ""}
        )
        last_event_ids_processed = [item.strip() for item in last_event_processed.value.split(",") if item]
        next_event_date: str = last_event_date_read.value
        events, readers, cardholders = get_sorted_reader_events(next_event_date, last_event_ids_processed)
        last_processed_event_index = None
        try:
            for i, event in enumerate(events):
                cardholder_id = event["CardholderId"]
                reader_id = event["SourceId"]
                event_time = event["EventDateObject"]
                reader = readers.get(reader_id)
                cardholder_sync_value = cardholders.get(cardholder_id, None)
                area = reader.area
                if area:
                    event_user = event.get("Cardholder")
                    user = find_user(cardholder_sync_value)
                    if user:
                        login_logout(user, reader, event_time, event)
                    else:
                        message = f"No NEMO matching user found for {rs2_settings.user_sync_field} {cardholder_sync_value} - {event_user} (event at location {reader.reader_name})"
                        add_error_log(
                            message,
                            ErrorLog.ErrorType.NO_MATCHING_USER,
                            cardholder_value=cardholder_sync_value,
                            event_data=event,
                        )
                # Only set next sequence number when we are done with the current event
                next_event_date = event["EventDate"]
                last_processed_event_index = i
        finally:
            if last_processed_event_index is not None:
                # Save the last processed event date and include the buffer
                next_event_date_time_end = datetime.datetime.fromisoformat(next_event_date).astimezone()
                next_event_date_time_start = next_event_date_time_end - datetime.timedelta(seconds=buffer_in_seconds)
                last_event_date_read.value = next_event_date_time_start.replace(tzinfo=None).isoformat(
                    sep="T", timespec="seconds"
                )
                last_event_date_read.save()
                # Use the buffer to find the last events between buffer and next_event_date
                new_last_event_ids_processed = []
                while (
                    last_processed_event_index >= 0
                    and next_event_date_time_start
                    <= events[last_processed_event_index]["EventDateObject"]
                    <= next_event_date_time_end
                ):
                    new_last_event_ids_processed.append(get_event_id(events[last_processed_event_index]))
                    last_processed_event_index -= 1
                last_event_processed.value = ",".join(new_last_event_ids_processed)
                last_event_processed.save()

    except Customization.DoesNotExist:
        message = (
            f"No last sequence number was set in Customization. Please add one with the name '{NEXT_EVENTS_BEGIN_DATE}'"
        )
        add_error_log(message)
        raise NEMOException(msg=message)
    except Exception as e:
        add_error_log(e)


def get_sorted_reader_events(
    begin_date: str, last_event_ids_processed: List[str]
) -> (List[Dict], Dict[int, Reader], Dict[int, str]):
    rs2_settings = get_rs2_settings()
    """
    Get events for all relevant readers
    Then merge them all (since login-logout happen in different readers)
    Finally we need to sort them all to recreate the chronological order of things
    """
    reader_ids = Reader.objects.filter(area__isnull=False).values_list("reader_id", flat=True)
    readers = Reader.objects.in_bulk(id_list=reader_ids, field_name="reader_id")
    cardholder_ids = set()
    cardholder_keys = dict()
    events = []
    for reader_id in readers:
        response = request_get(READERS_HISTORY_URL.format(reader_id, begin_date))
        response.raise_for_status()
        reader_events = response.json()
        for reader_event in reader_events:
            # Skip events that are not relevant
            if reader_event["EventType"] in rs2_settings.relevant_event_types:
                reader_event["SourceId"] = reader_id
                # Skip events we already processed
                if get_event_id(reader_event) not in last_event_ids_processed:
                    reader_event["EventDateObject"] = datetime.datetime.fromisoformat(
                        reader_event["EventDate"]
                    ).astimezone()
                    events.append(reader_event)
                    cardholder_id = reader_event.get("CardholderId", None)
                    if cardholder_id:
                        # If we need to sync cardholder objects, store the id for later use
                        if rs2_settings.cardholder_sync:
                            cardholder_ids.add(cardholder_id)
                        # Otherwise we can just grab the field from the event
                        else:
                            cardholder_keys[cardholder_id] = reader_event.get(rs2_settings.user_sync_field)
    events.sort(key=lambda x: x["EventDateObject"])
    # Now we can deal with cardholders in bulk
    if cardholder_ids:
        key_name = rs2_settings.user_sync_field
        # Grab the ones we already synced before
        for cardholder in Cardholder.objects.filter(cardholder_id__in=cardholder_ids, key_name=key_name):
            cardholder_keys[cardholder.cardholder_id] = cardholder.key_value
        # Fetch the missing ones
        for missing_cardholder_id in cardholder_ids.difference(cardholder_keys.keys()):
            cardholder_keys[missing_cardholder_id] = get_cardholder_property(missing_cardholder_id, key_name)
    return events, readers, cardholder_keys


def get_cardholder_property(cardholder_id: str, key_name, save=True) -> Optional[str]:
    # Request cardholder data from the API and save it, returning the value
    response = request_get(CARDHOLDERS_URL.format(cardholder_id))
    response.raise_for_status()
    rs2_cardholder = response.json()[0]
    value: str = find_in_json(key_name, rs2_cardholder)
    cardholder_name = rs2_cardholder.get("FirstName") + " " + rs2_cardholder.get("LastName")
    if value:
        value = value.strip()
        if save:
            Cardholder.objects.update_or_create(
                cardholder_id=cardholder_id,
                defaults={"cardholder_name": cardholder_name, "key_name": key_name, "key_value": value},
            )
        return value


def sync_readers():
    sync_model_rest(READERS_URL, Reader, "ReaderID", reader_mapping)


def sync_model_rest(url, model_class: Type[Model], remote_id_field_name: str, mapping: Dict):
    sync_model_rest_data(request_get(url).json(), model_class, remote_id_field_name, mapping)


def request_get(url_suffix: str):
    rs2_settings = get_rs2_settings()
    if rs2_settings:
        auth = (
            HTTPBasicAuth(rs2_settings.auth.get("user_id"), rs2_settings.auth.get("password"))
            if getattr(rs2_settings, "auth")
            else None
        )
        headers = {"PublicKey": rs2_settings.public_key}
        timeout = getattr(rs2_settings, "timeout", 30)
        return requests.get(rs2_settings.url + url_suffix, auth=auth, headers=headers, timeout=timeout)
    else:
        return HttpResponseBadRequest("no RS2_ACCESS settings found, please add them to your settings.py")


def login_logout(user: User, reader: Reader, event_time: datetime.datetime = None, event=None):
    if reader.reader_type == Reader.ReaderType.ENTRANCE:
        login(user, reader.area, event_time, event)
    elif reader.reader_type == Reader.ReaderType.EXIT:
        logout(user, event_time, event)


def login(user: User, area: Area, start: datetime.datetime = None, event=None):
    try:
        default_project = user.get_preferences().default_project.default_project
    except UserPreferencesDefaultProject.DoesNotExist:
        default_project = None
    user_active_projects = user.active_projects()
    # Use the user's default project if it exists and is active, otherwise use the first active project
    if default_project and default_project in user_active_projects:
        project = default_project
    else:
        project = user_active_projects.first()
    if project:
        area_access: AreaAccessRecord = user.area_access_record()
        # if user is already logged in the same area, record it in error log
        if area_access and area_access.area == area:
            if get_rs2_settings().record_already_logged_in_as_error:
                message = f"user {user} attempted to login but was already logged in to the {area}"
                add_error_log(message, ErrorLog.ErrorType.ALREADY_LOGGED_IN, user=user, event_data=event)
        # If user is logged in a different area, automatically log him out
        if area_access and area_access.area != area:
            logout(user, start, event)
        # Only log in if not already in the area
        if not area_access or area_access.area != area:
            AreaAccessRecord.objects.create(area=area, customer=user, start=start, project=project)
    else:
        message = f"no active projects found for user {user}, skipping log in to {area}"
        add_error_log(message, ErrorLog.ErrorType.NO_ACTIVE_PROJECTS, user=user, event_data=event)


def logout(user: User, end: datetime.datetime, event=None):
    area_access: AreaAccessRecord = user.area_access_record()
    if area_access:
        for project_id in get_project_ids_to_charge_for_user(user, area_access.start, end, area_access.project_id):
            refresh_access: AreaAccessRecord = AreaAccessRecord.objects.get(pk=area_access.id)
            refresh_access.project_id = project_id
            if not refresh_access.end:
                # update current record
                refresh_access.end = end
                refresh_access.save()
            else:
                # We are creating a copy for each project the user worked on
                new_access = new_model_copy(refresh_access)
                new_access.save()

        # Dealing with reservation and staff charges, to be consistent with NEMO
        shorten_reservation(user, area_access.area, end)
        # Stop charging area access if staff is leaving the area
        staff_charge = user.get_staff_charge()
        if staff_charge:
            try:
                staff_area_access = AreaAccessRecord.objects.get(staff_charge=staff_charge, end=None)
                staff_area_access.end = timezone.now()
                staff_area_access.save()
            except AreaAccessRecord.DoesNotExist:
                pass
    else:
        message = f"user {user} logged out but was not previously logged in to any areas"
        add_error_log(message, ErrorLog.ErrorType.LOGOUT_WITHOUT_RECORD, user=user, event_data=event)


def get_project_ids_to_charge_for_user(
    user: User, start: datetime.datetime, end: datetime.datetime, original_project_id: int
) -> Union[List, Set]:
    """
    Figure out which project to charge for the user, given a datetime
    1. If the user has no active projects, use the same project as for login (unlikely, but possible)
    2. If the user has only one active project, use it
    3. Be smart and check tool usage for this user (non-remote) charge once per account
    4. No tool usage, use default project
    5. No default project, use the first active project by relation id (== first one added)
    """
    # We are using the through relation to get order by relation id, which means ordered by added time (lower id)
    active_through_projects = user.projects.through.objects.filter(
        user_id=user.id, project__active=True, project__account__active=True
    )
    active_through_projects_count = active_through_projects.count()
    # Case #1
    if active_through_projects_count == 0:
        message = (
            f"no active projects found for user {user}, using original project id {original_project_id} for logout"
        )
        add_error_log(message, ErrorLog.ErrorType.NO_ACTIVE_PROJECTS, user=user)
        return [original_project_id]
    # Case #2
    if active_through_projects_count == 1:
        return list(active_through_projects.values_list("project_id", flat=True))
    # Case #3
    ongoing_events = (
        UsageEvent.objects.filter(user=user, end__isnull=True)
        .exclude(remote_work=True)
        .exclude(start__lt=start)
        .exclude(start__gt=end)
    )
    other_events = (
        UsageEvent.objects.filter(user=user, end__isnull=False)
        .exclude(remote_work=True)
        .exclude(start__lt=start, end__lt=start)
        .exclude(start__gt=end, end__gt=end)
    )
    project_ids = list(
        distinct_qs_value_list(ongoing_events, "project_id") | distinct_qs_value_list(other_events, "project_id")
    )
    if project_ids:
        unique_account_project_ids = []
        projects = Project.objects.in_bulk(project_ids)
        already_checked_account_ids = set()
        for project_id, project in projects.items():
            if project.account_id not in already_checked_account_ids:
                unique_account_project_ids.append(project_id)
                already_checked_account_ids.add(project.account_id)
        return unique_account_project_ids
    else:
        # Case #4
        try:
            default_project_id = user.get_preferences().default_project.default_project_id
            if default_project_id:
                if default_project_id in user.active_projects().values_list("id", flat=True):
                    return [default_project_id]
                else:
                    user.get_preferences().default_project.delete()
                    user_office_email = EmailsCustomization.get("user_office_email_address")
                    content = f"Dear {user.get_name()},<br><br>"
                    content += f"Your default project was reset because you are not a member of it anymore.<br><br>"
                    content += f'Please go to <a href="{get_full_url(reverse("user_preferences"))}">you preferences page</a> and select a new one.'
                    send_mail(
                        subject=f"Default project reset",
                        content=content,
                        from_email=None,
                        to=user.get_emails(EmailNotificationType.BOTH_EMAILS),
                        cc=[user_office_email],
                    )
        except:
            pass
        # Case #5
        return [active_through_projects.first().project_id]


def update_current_access_project():
    for area_access_record in AreaAccessRecord.objects.filter(end__isnull=True):
        project_id_to_charge = get_project_ids_to_charge_for_user(
            area_access_record.customer, area_access_record.start, timezone.now(), area_access_record.project_id
        )
        if project_id_to_charge and project_id_to_charge != area_access_record.project_id:
            area_access_record.project_id = project_id_to_charge[0]
            area_access_record.save(update_fields=["project_id"])


def add_error_log(
    description: Union[str, Exception],
    error_type: ErrorLog.ErrorType = ErrorLog.ErrorType.GENERAL,
    user=None,
    cardholder_value=None,
    event_data: Dict = None,
):
    if event_data:
        # remove datetime object which is not serializable
        event_data = {key: value for key, value in event_data.items() if key != "EventDateObject"}
    # Force description to str
    ErrorLog.objects.create(
        description=str(description),
        error_type=error_type,
        cardholder_value=cardholder_value,
        user=user,
        event_data=json.dumps(event_data) if event_data else None,
    )
    if isinstance(description, Exception):
        rs2_logger.exception(str(description))
    else:
        rs2_logger.error(description)



# Generated by Django 3.2.20 on 2023-08-10 18:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('NEMO_rs2_access', '0003_errorlog'),
    ]

    operations = [
        migrations.RenameField(
            model_name='errorlog',
            old_name='data',
            new_name='cardholder_value',
        ),
        migrations.AddField(
            model_name='errorlog',
            name='event_data',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='errorlog',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='errorlog',
            name='error_type',
            field=models.PositiveIntegerField(choices=[(1, 'General'), (2, 'No active projects'), (3, 'No matching user'), (4, 'Logout without record')], default=1),
        ),
    ]
